# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import sys

import pytest

from idf_ci.utils import remove_subfolders


@pytest.mark.parametrize(
    'relpaths,expected_relpaths',
    [
        (['a/b/c', 'a/b', 'a', 'a/b/c/d'], ['a']),
        (['b', 'a/b', 'a', 'c/d'], ['a', 'b', 'c/d']),
    ],
)
@pytest.mark.skipif(sys.platform == 'win32', reason='Using Unix paths')
def test_remove_subfolders(tmp_path, relpaths, expected_relpaths):
    paths = []
    for rel in relpaths:
        dir_path = tmp_path / rel
        dir_path.mkdir(parents=True, exist_ok=True)
        paths.append(str(dir_path))
    expected = [str(tmp_path / rel) for rel in expected_relpaths]
    assert remove_subfolders(paths) == expected



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
import tempfile

import pytest

from idf_ci.idf_gitlab import ArtifactParams


@pytest.fixture(autouse=True)
def setup(tmp_path):
    """Change to temporary directory for tests."""
    curdir = os.getcwd()
    os.chdir(tmp_path)

    yield

    os.chdir(curdir)


@pytest.fixture(autouse=True)
def setup_git_repo(tmp_path):
    """Set up a git repository with initial commit and optional branch."""
    # Create a git repo
    subprocess.run(['git', 'init'], check=True)
    subprocess.run(['git', 'config', 'user.email', 'test@example.com'], check=True)
    subprocess.run(['git', 'config', 'user.name', 'Test User'], check=True)

    # Create initial commit on feature branch
    subprocess.run(['git', 'checkout', '-b', 'feature-branch'], check=True)
    (tmp_path / 'test.txt').write_text('test')
    subprocess.run(['git', 'add', 'test.txt'], check=True)
    subprocess.run(['git', 'commit', '-m', 'feature commit'], check=True)

    # Create local branch
    subprocess.run(['git', 'checkout', '-b', 'local-branch'], check=True)
    (tmp_path / 'local.txt').write_text('local content')
    subprocess.run(['git', 'add', 'local.txt'], check=True)
    subprocess.run(['git', 'commit', '-m', 'local commit'], check=True)

    return tmp_path


def get_commit_sha(ref='HEAD'):
    """Get commit SHA for a given reference."""
    return subprocess.run(
        ['git', 'rev-parse', ref],
        check=True,
        capture_output=True,
        text=True,
    ).stdout.strip()


@pytest.fixture
def feature_sha():
    """Set up git repo with feature branch and return its SHA."""
    return get_commit_sha('feature-branch')


@pytest.fixture
def local_sha():
    """Set up git repo with local branch and return its SHA."""
    return get_commit_sha('local-branch')


def test_commit_sha_precedence(monkeypatch, feature_sha, local_sha):
    """Test the precedence of commit SHA resolution in StorageParams."""
    # 1. Test CLI commit SHA has the highest priority
    params = ArtifactParams(commit_sha='cli-sha')
    assert params.commit_sha == 'cli-sha'

    # 2. Test PIPELINE_COMMIT_SHA has second priority
    monkeypatch.setenv('PIPELINE_COMMIT_SHA', 'env-sha')
    params = ArtifactParams()
    assert params.commit_sha == 'env-sha'

    # 3. Test specified branch precedes local branch
    monkeypatch.delenv('PIPELINE_COMMIT_SHA')
    params = ArtifactParams(branch='feature-branch')
    assert params.commit_sha == feature_sha

    # 4. Test local branch latest commit
    params = ArtifactParams()
    assert params.commit_sha == local_sha

    # 5. Test ValueError when no commit SHA can be determined
    new_empty_folder = os.path.join(tempfile.gettempdir(), 'new_empty_folder')
    os.makedirs(new_empty_folder, exist_ok=True)
    os.chdir(new_empty_folder)

    with pytest.raises(
        ValueError,
        match='Failed to get commit SHA from git command. '
        'Must set commit_sha or branch parameter, or set PIPELINE_COMMIT_SHA env var',
    ):
        ArtifactParams()



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
"""This file is used for generating the child pipeline for build jobs."""

import logging
import os
import typing as t

import yaml
from idf_build_apps import App
from jinja2 import Environment

from idf_ci.envs import GitlabEnvVars
from idf_ci.idf_pytest import GroupedPytestCases, get_pytest_cases
from idf_ci.scripts import get_all_apps
from idf_ci.settings import CiSettings

logger = logging.getLogger(__name__)


def _get_fake_pass_job(settings: CiSettings, workflow_name: str) -> t.Dict[str, t.Any]:
    # no matter being used in build or test child pipeline,
    # always use the same fake_pass job that extends the build job template
    # since test child pipeline tags are generated programmatically
    return {
        'fake_pass': {
            'stage': settings.gitlab.build_pipeline.job_stage,
            'tags': settings.gitlab.build_pipeline.job_tags,
            'before_script': [],
            'after_script': [],
            'cache': [],
            'needs': [],
            'script': [
                'echo "skip the entire child pipeline"',
            ],
        },
        # required for GitLab to recognize this as a valid pipeline
        'workflow': {
            'name': workflow_name,  # different workflow names for build and test child pipelines
            'rules': [
                {'when': 'always'},
            ],
        },
    }


def dump_apps_to_txt(apps: t.List[App], output_file: str) -> None:
    """Dump a list of apps to a text file, one app per line."""
    with open(output_file, 'w') as fw:
        for app in apps:
            fw.write(app.model_dump_json() + '\n')


def build_child_pipeline(
    *,
    paths: t.Optional[t.List[str]] = None,
    modified_files: t.Optional[t.List[str]] = None,
    compare_manifest_sha_filepath: t.Optional[str] = None,
    yaml_output: t.Optional[str] = None,
) -> None:
    """Generate build child pipeline."""
    envs = GitlabEnvVars()
    settings = CiSettings()

    if compare_manifest_sha_filepath and not os.path.isfile(compare_manifest_sha_filepath):
        compare_manifest_sha_filepath = None

    if yaml_output is None:
        yaml_output = settings.gitlab.build_pipeline.yaml_filename

    # Check if we should run quick pipeline
    if envs.select_by_filter_expr:
        # we only build test related apps
        test_related_apps, _ = get_all_apps(
            paths=paths,
            marker_expr='not host_test',
            filter_expr=envs.select_by_filter_expr,
        )
        non_test_related_apps: t.List[App] = []
        dump_apps_to_txt(test_related_apps, settings.collected_test_related_apps_filepath)
    else:
        test_related_apps, non_test_related_apps = get_all_apps(
            paths=paths,
            modified_files=modified_files,
            marker_expr='not host_test',
            compare_manifest_sha_filepath=compare_manifest_sha_filepath,
        )
        dump_apps_to_txt(test_related_apps, settings.collected_test_related_apps_filepath)
        dump_apps_to_txt(non_test_related_apps, settings.collected_non_test_related_apps_filepath)

    apps_total = len(test_related_apps) + len(non_test_related_apps)
    test_related_parallel_count = (
        len(test_related_apps) // settings.gitlab.build_pipeline.runs_per_job + 1 if test_related_apps else 0
    )
    non_test_related_parallel_count = (
        len(non_test_related_apps) // settings.gitlab.build_pipeline.runs_per_job + 1 if non_test_related_apps else 0
    )

    if not apps_total:
        logger.info('No apps found, generating fake_pass job to skip the entire build child pipeline')
        with open(yaml_output, 'w') as fw:
            yaml.safe_dump(_get_fake_pass_job(settings, settings.gitlab.build_pipeline.workflow_name), fw)
            return

    logger.info(
        'Found %d apps, %d test related apps, %d non-test related apps',
        apps_total,
        len(test_related_apps),
        len(non_test_related_apps),
    )
    logger.info(
        'Test related parallel count: %d, Non-test related parallel count: %d',
        test_related_parallel_count,
        non_test_related_parallel_count,
    )

    job_template = Environment().from_string(settings.gitlab.build_pipeline.job_template_jinja)
    jobs_template = Environment().from_string(settings.gitlab.build_pipeline.jobs_jinja)
    yaml_template = Environment().from_string(settings.gitlab.build_pipeline.yaml_jinja)

    with open(yaml_output, 'w') as fw:
        fw.write(
            yaml_template.render(
                job_template=job_template.render(
                    settings=settings,
                ),
                jobs=jobs_template.render(
                    settings=settings,
                    test_related_apps_count=len(test_related_apps),
                    test_related_parallel_count=test_related_parallel_count,
                    non_test_related_apps_count=len(non_test_related_apps),
                    non_test_related_parallel_count=non_test_related_parallel_count,
                ),
                settings=settings,
            )
        )

    logger.info('Build child pipeline generated successfully in %s', yaml_output)


def test_child_pipeline(
    yaml_output: str,
    *,
    cases: t.Optional[GroupedPytestCases] = None,
) -> None:
    """This function is used to generate the child pipeline for test jobs.

    Suppose the ci_build_artifacts_filepatterns is downloaded already

    .. note::

        parallel:matrix does not support array as value, we generate all jobs here

    Example output:

    .. code-block:: yaml

        .default_test_settings:
            script:
                - pytest ${nodes}

        esp32 - generic:
            extends:
                - .default_test_settings
            tags:
                - esp32
                - generic
            variables:
                nodes: "nodeid1 nodeid2"
    """
    settings = CiSettings()

    if yaml_output is None:
        yaml_output = settings.gitlab.test_pipeline.yaml_filename

    if cases is None:
        cases = GroupedPytestCases(get_pytest_cases())

    if not cases.grouped_cases:
        logger.info('No test cases found, generating fake_pass job to skip the entire test child pipeline')
        with open(yaml_output, 'w') as fw:
            yaml.safe_dump(_get_fake_pass_job(settings, settings.gitlab.test_pipeline.workflow_name), fw)
        return

    jobs = []
    for key, grouped_cases in cases.grouped_cases.items():
        jobs.append(
            {
                'name': f'{key.target_selector} - {key.env_selector}',
                'tags': sorted(key.runner_tags),
                # quote nodeids to avoid special chars issues
                'nodes': '"' + ' '.join([f"'{c.item.nodeid}'" for c in grouped_cases]) + '"',
                'parallel_count': len(grouped_cases) // settings.gitlab.test_pipeline.runs_per_job + 1,
                **cases.additional_dict.get(key, {}),
            }
        )

    job_template = Environment().from_string(settings.gitlab.test_pipeline.job_template_jinja)
    jobs_template = Environment().from_string(settings.gitlab.test_pipeline.jobs_jinja)
    yaml_template = Environment().from_string(settings.gitlab.test_pipeline.yaml_jinja)

    with open(yaml_output, 'w') as fw:
        fw.write(
            yaml_template.render(
                default_template=job_template.render(
                    settings=settings,
                ),
                jobs=jobs_template.render(
                    jobs=jobs,
                    settings=settings,
                ),
                settings=settings,
            )
        )

    logger.info('Test child pipeline generated successfully in %s', yaml_output)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import re
import sys
import typing as t
from pathlib import Path

from idf_build_apps import App, json_list_files_to_apps
from idf_build_apps.constants import BuildStatus
from pydantic_settings import (
    BaseSettings,
    InitSettingsSource,
    PydanticBaseSettingsSource,
)
from tomlkit import load

from idf_ci._compat import PathLike, TypedDict

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired


logger = logging.getLogger(__name__)


class TomlConfigSettingsSource(InitSettingsSource):
    """A source class that loads variables from a TOML file"""

    def __init__(
        self,
        settings_cls: t.Type[BaseSettings],
        toml_file: t.Optional[PathLike] = Path(''),
    ):
        self.toml_file_path = self._pick_toml_file(
            toml_file,
            '.idf_ci.toml',
        )
        self.toml_data = self._read_file(self.toml_file_path)
        super().__init__(settings_cls, self.toml_data)

    def _read_file(self, path: t.Optional[Path]) -> t.Dict[str, t.Any]:
        if not path or not path.is_file():
            return {}

        with open(path) as f:
            return load(f)

    @staticmethod
    def _pick_toml_file(provided: t.Optional[PathLike], filename: str) -> t.Optional[Path]:
        """Pick a file path to use.

        If a file path is provided, use it. Otherwise, search up the directory tree for
        a file with the given name.

        :param provided: Explicit path provided when instantiating this class.
        :param filename: Name of the file to search for.
        """
        if provided:
            provided_p = Path(provided)
            if provided_p.is_file():
                fp = provided_p.resolve()
                logger.debug(f'Loading config file: {fp}')
                return fp

        rv = Path.cwd()
        while len(rv.parts) > 1:
            fp = rv / filename
            if fp.is_file():
                logger.debug(f'Loading config file: {fp}')
                return fp

            rv = rv.parent

        return None


class S3FilePatternConfig(TypedDict):
    bucket: str
    patterns: t.List[str]
    if_clause: NotRequired[str]
    """List of glob patterns for files to collect."""


class ArtifactSettings(BaseSettings):
    ### in s3 buckets ###
    s3: t.Dict[str, S3FilePatternConfig] = {
        'debug': {
            'bucket': 'idf-artifacts',
            'patterns': [
                '**/build*/bootloader/*.map',
                '**/build*/bootloader/*.elf',
                '**/build*/*.map',
                '**/build*/*.elf',
                '**/build*/build.log',  # build_log_filename
            ],
        },
        'flash': {
            'bucket': 'idf-artifacts',
            'patterns': [
                '**/build*/bootloader/*.bin',
                '**/build*/*.bin',
                '**/build*/partition_table/*.bin',
                '**/build*/flasher_args.json',
                '**/build*/flash_project_args',
                '**/build*/config/sdkconfig.json',
                '**/build*/sdkconfig',
                '**/build*/project_description.json',
            ],
        },
    }
    """Dictionary mapping artifact types to their bucket and file patterns."""

    ### not in s3 buckets ###

    build_job_filepatterns: t.List[str] = [
        'app_info_*.txt',  # collect_app_info_filename
        'build_summary_*.xml',  # junitxml
    ]
    """List of glob patterns for CI build jobs artifacts to collect."""

    test_job_filepatterns: t.List[str] = [
        'pytest-embedded/',
        'XUNIT_RESULT*.xml',
    ]
    """List of glob patterns for CI test jobs artifacts to collect."""

    @property
    def available_s3_types(self) -> t.List[str]:
        """Get list of available S3 artifact types.

        :returns: List of artifact type names
        """
        return sorted(self.s3.keys())


class BuildPipelineSettings(BaseSettings):
    workflow_name: str = 'Build Child Pipeline'
    """Name for the GitLab CI workflow."""

    presigned_json_job_name: str = 'generate_presigned_json'
    """Name of the job within the child pipeline that creates the presigned.json file."""

    job_template_name: str = '.default_build_settings'
    """Default template name for CI build jobs."""

    job_stage: str = 'build'
    """Default stage for build jobs in the pipeline."""

    job_template_jinja: str = """
{{ settings.gitlab.build_pipeline.job_template_name }}:
  stage: {{ settings.gitlab.build_pipeline.job_stage }}
  tags: {{ settings.gitlab.build_pipeline.job_tags }}
  timeout: 1h
  artifacts:
    paths:
    {%- for path in settings.gitlab.artifacts.build_job_filepatterns %}
      - {{ path }}
    {%- endfor %}
    expire_in: 1 week
    when: always
  script:
    - idf-ci build run
      --parallel-count ${CI_NODE_TOTAL:-1}
      --parallel-index ${CI_NODE_INDEX:-1}
""".strip()
    """Default template for CI build jobs."""

    job_tags: t.List[str] = ['build']
    """List of tags for CI build jobs."""

    runs_per_job: int = 60
    """Maximum number of apps to build in a single job."""

    jobs_jinja: str = """
{%- if test_related_apps_count > 0 %}
build_test_related_apps:
  extends: {{ settings.gitlab.build_pipeline.job_template_name }}
{%- if test_related_parallel_count > 1 %}
  parallel: {{ test_related_parallel_count }}
{%- endif %}
  needs:
    - pipeline: $PARENT_PIPELINE_ID
      job: generate_build_child_pipeline
    - pipeline: $PARENT_PIPELINE_ID
      job: pipeline_variables
  variables:
    IDF_CI_BUILD_ONLY_TEST_RELATED_APPS: "1"
{%- endif %}

{%- if non_test_related_apps_count > 0 %}
build_non_test_related_apps:
  extends: {{ settings.gitlab.build_pipeline.job_template_name }}
{%- if non_test_related_parallel_count > 1 %}
  parallel: {{ non_test_related_parallel_count }}
{%- endif %}
  needs:
    - pipeline: $PARENT_PIPELINE_ID
      job: generate_build_child_pipeline
    - pipeline: $PARENT_PIPELINE_ID
      job: pipeline_variables
  variables:
    IDF_CI_BUILD_ONLY_NON_TEST_RELATED_APPS: "1"
{%- endif %}
""".strip()
    """Jinja2 template for build jobs configuration."""

    pre_yaml_jinja: str = ''
    """yaml content to be injected before the yaml template."""

    yaml_jinja: str = """
{{ settings.gitlab.build_pipeline.pre_yaml_jinja }}

workflow:
  name: {{ settings.gitlab.build_pipeline.workflow_name }}
  rules:
    - when: always

{%- if settings.gitlab.build_pipeline.job_template_jinja %}
{{ job_template }}
{%- endif %}

{{ jobs }}

{%- if test_related_apps_count > 0 %}
generate_test_child_pipeline:
  extends: {{ settings.gitlab.build_pipeline.job_template_name }}
  needs:
    - build_test_related_apps
  artifacts:
    paths:
      - {{ settings.gitlab.test_pipeline.yaml_filename }}
  script:
    - idf-ci gitlab test-child-pipeline

test-child-pipeline:
  stage: .post
  needs:
    - generate_test_child_pipeline
  variables:
    PARENT_PIPELINE_ID: $PARENT_PIPELINE_ID
  trigger:
    include:
      - artifact: {{ settings.gitlab.test_pipeline.yaml_filename }}
        job: generate_test_child_pipeline
    strategy: depend
{%- endif %}
""".strip()
    """Jinja2 template for the build child pipeline YAML content."""

    yaml_filename: str = 'build_child_pipeline.yml'
    """Filename for the build child pipeline YAML file."""


class TestPipelineSettings(BuildPipelineSettings):
    workflow_name: str = 'Test Child Pipeline'
    """Name for the GitLab CI workflow."""

    job_template_name: str = '.default_test_settings'
    """Default template name for CI test jobs."""

    job_stage: str = 'test'
    """Default stage for test jobs in the pipeline."""

    job_template_jinja: str = """
{{ settings.gitlab.test_pipeline.job_template_name }}:
  stage: {{ settings.gitlab.test_pipeline.job_stage }}
  timeout: 1h
  artifacts:
    paths:
    {%- for path in settings.gitlab.artifacts.test_job_filepatterns %}
      - {{ path }}
    {%- endfor %}
    expire_in: 1 week
    when: always
  variables:
    PYTEST_EXTRA_FLAGS: ""
  script:
    - pytest ${nodes}
      --parallel-count ${CI_NODE_TOTAL:-1}
      --parallel-index ${CI_NODE_INDEX:-1}
      --junitxml XUNIT_RESULT_${CI_JOB_NAME_SLUG}.xml
      ${PYTEST_EXTRA_FLAGS}
""".strip()
    """Default template for CI test jobs."""

    job_tags: t.List[str] = []
    """Unused. tags are set by test cases."""

    runs_per_job: int = 30
    """Maximum number of test cases to run in a single job."""

    jobs_jinja: str = """
{% for job in jobs %}
{{ job['name'] }}:
  extends:
    - {{ settings.gitlab.test_pipeline.job_template_name }}
    {%- for extra_extend in job.get('extra_extends', []) %}
    - {{ extra_extend }}
    {%- endfor %}
  tags: {{ job['tags'] }}
{%- if job['parallel_count'] > 1 %}
  parallel: {{ job['parallel_count'] }}
{%- endif %}
  variables:
    nodes: {{ job['nodes'] }}
{% endfor %}
""".strip()
    """Jinja2 template for test jobs configuration."""

    yaml_jinja: str = """
{{ settings.gitlab.test_pipeline.pre_yaml_jinja }}

workflow:
  name: {{ settings.gitlab.test_pipeline.workflow_name }}
  rules:
    - when: always

{%- if settings.gitlab.test_pipeline.job_template_jinja %}
{{ default_template }}
{%- endif %}

{{ jobs }}
""".strip()
    """Jinja2 template for the test child pipeline YAML content."""

    yaml_filename: str = 'test_child_pipeline.yml'
    """Filename for the test child pipeline YAML file."""


class GitlabSettings(BaseSettings):
    project: str = 'espressif/esp-idf'
    """GitLab project path in the format 'owner/repo'."""

    known_failure_cases_bucket_name: str = 'ignore-test-result-files'
    """Bucket name for storing known failure cases."""

    artifacts: ArtifactSettings = ArtifactSettings()

    build_pipeline: BuildPipelineSettings = BuildPipelineSettings()

    test_pipeline: TestPipelineSettings = TestPipelineSettings()


class CiSettings(BaseSettings):
    CONFIG_FILE_PATH: t.ClassVar[t.Optional[Path]] = None
    """Path to the configuration file to be used (class variable)."""

    component_mapping_regexes: t.List[str] = [
        '/components/(.+?)/',
        '/common_components/(.+?)/',
    ]
    """List of regex patterns to extract component names from file paths."""

    extend_component_mapping_regexes: t.List[str] = []
    """Additional component mapping regex patterns to extend the default list."""

    component_mapping_exclude_regexes: t.List[str] = [
        r'/test_apps/',
    ]
    """List of regex patterns to exclude certain paths from component mapping."""

    component_ignored_file_extensions: t.List[str] = [
        '.md',
        '.rst',
        '.yaml',
        '.yml',
        '.py',
    ]
    """File extensions to ignore when mapping files to components."""

    extend_component_ignored_file_extensions: t.List[str] = []
    """Additional file extensions to ignore."""

    # build related settings
    built_app_list_filepatterns: t.List[str] = ['app_info_*.txt']
    """Glob patterns for files containing built app information."""

    collected_test_related_apps_filepath: str = 'test_related_apps.txt'
    """Path to file containing test-related apps."""

    collected_non_test_related_apps_filepath: str = 'non_test_related_apps.txt'
    """Path to file containing non-test-related apps."""

    preserve_test_related_apps: bool = True
    """Whether to preserve test-related apps."""

    preserve_non_test_related_apps: bool = True
    """Whether to preserve non-test-related apps."""

    extra_default_build_targets: t.List[str] = []
    """Additional build targets to include by default."""

    # env vars
    ci_detection_envs: t.List[str] = [
        'CI',
        'GITHUB_ACTIONS',
        'CIRCLECI',
        'TRAVIS',
        'JENKINS_URL',
        'DRONE',
        'APPVEYOR',
        'BITBUCKET_COMMIT',
        'SEMAPHORE',
        'TEAMCITY_VERSION',
    ]
    """Environment variables used to detect if running in CI."""

    local_runtime_envs: t.Dict[str, t.Any] = {}
    """Environment variables to set in local development."""

    ci_runtime_envs: t.Dict[str, t.Any] = {}
    """Environment variables to set in CI environment."""

    # gitlab subsection
    gitlab: GitlabSettings = GitlabSettings()
    """GitLab-specific settings."""

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: t.Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        dotenv_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        file_secret_settings: PydanticBaseSettingsSource,  # noqa: ARG003
    ) -> t.Tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            TomlConfigSettingsSource(
                settings_cls, cls.CONFIG_FILE_PATH if cls.CONFIG_FILE_PATH is not None else '.idf_ci.toml'
            ),
        )

    def model_post_init(self, __context: t.Any) -> None:
        runtime_envs = self.ci_runtime_envs if self.is_in_ci else self.local_runtime_envs

        for key, value in runtime_envs.items():
            os.environ[key] = str(value)
            logger.debug('Set local env var: %s=%s', key, value)

    @property
    def is_in_ci(self) -> bool:
        """Check if the code is running in a CI environment.

        :returns: True if in CI environment, False otherwise
        """
        return any(os.getenv(env) is not None for env in self.ci_detection_envs)

    @property
    def all_component_mapping_regexes(self) -> t.Set[re.Pattern]:
        """Get all component mapping regexes as compiled pattern objects.

        :returns: Set of compiled regex patterns
        """
        return {re.compile(regex) for regex in self.component_mapping_regexes + self.extend_component_mapping_regexes}

    @property
    def all_component_mapping_exclude_regexes(self) -> t.Set[re.Pattern]:
        """Get all component mapping exclude regexes as compiled pattern objects.

        :returns: Set of compiled regex patterns
        """
        return {re.compile(regex) for regex in self.component_mapping_exclude_regexes}

    def get_modified_components(self, modified_files: t.Iterable[str]) -> t.Set[str]:
        """Get the set of components that have been modified based on the provided files.

        :param modified_files: Iterable of file paths that have been modified

        :returns: Set of component names that have been modified
        """
        modified_components = set()

        for modified_file in modified_files:
            file_path = Path(modified_file)
            if (
                file_path.suffix
                in self.component_ignored_file_extensions + self.extend_component_ignored_file_extensions
            ):
                continue

            # always use absolute path as posix string
            # Path.resolve return relative path when file does not exist. so use os.path.abspath
            abs_path = Path(os.path.abspath(modified_file)).as_posix()

            for regex in self.all_component_mapping_regexes:
                match = regex.search(abs_path)
                if match:
                    for exclude_regex in self.all_component_mapping_exclude_regexes:
                        if exclude_regex.search(abs_path):
                            logger.debug(f'Excluding {abs_path} from component mapping')
                            break
                    else:
                        modified_components.add(match.group(1))
                        break

        return modified_components

    @classmethod
    def read_apps_from_files(cls, filepaths: t.Sequence[PathLike]) -> t.Optional[t.List[App]]:
        """Helper method to read apps from files.

        :param filepaths: List of file paths to read

        :returns: List of App objects read from the files, or None if no files found
        """
        valid_filepaths = []
        for filepath in filepaths:
            if os.path.isfile(filepath):
                valid_filepaths.append(str(filepath))
            else:
                logger.debug(f'No file found: {filepath}')

        if not valid_filepaths:
            return None

        return json_list_files_to_apps(valid_filepaths)

    @classmethod
    def read_apps_from_filepatterns(cls, patterns: t.List[str]) -> t.Optional[t.List[App]]:
        """Helper method to read apps from files matching given patterns.

        :param patterns: List of file patterns to search for

        :returns: List of App objects read from the files, or None if no files found
        """
        found_files = []
        for filepattern in patterns:
            found_files.extend([str(p) for p in (Path('.').glob(filepattern))])

        if not found_files:
            logger.debug(f'No files found for patterns: {patterns}')
            return None

        apps = json_list_files_to_apps(found_files)

        if not apps:
            logger.warning(f'No apps found for patterns: {patterns}, returning empty list')

        return apps

    def get_built_apps_list(self) -> t.Optional[t.List[App]]:
        """Get the list of successfully built applications from the app info files.

        :returns: List of App objects representing successfully built applications, or
            None if no files found
        """
        # Read apps from files
        apps = self.read_apps_from_filepatterns(self.built_app_list_filepatterns)

        if apps is None:
            return None

        # Filter for successful builds
        built_apps = [app for app in apps if app.build_status == BuildStatus.SUCCESS]

        return built_apps



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import importlib.util
import logging
import os
import re
import sys
import typing as t
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.config import Config
from _pytest.fixtures import FixtureRequest
from _pytest.python import Function
from _pytest.stash import StashKey
from pytest_embedded.plugin import multi_dut_argument, multi_dut_fixture

from ..settings import CiSettings
from ..utils import setup_logging
from .models import PytestCase

_MODULE_NOT_FOUND_REGEX = re.compile(r"No module named '(.+?)'")
IDF_CI_PYTEST_CASE_KEY = StashKey[t.Optional[PytestCase]]()
IDF_CI_PYTEST_DEBUG_INFO_KEY = StashKey[t.Dict[str, t.Any]]()
IDF_CI_PLUGIN_KEY = StashKey['IdfPytestPlugin']()

logger = logging.getLogger(__name__)


##########
# Plugin #
##########
class IdfPytestPlugin:
    def __init__(
        self,
        *,
        cli_target: str,
        sdkconfig_name: t.Optional[str] = None,
    ) -> None:
        """Initialize the IDF pytest plugin.

        :param cli_target: Target passed from command line - single target, comma
            separated targets, or 'all'
        :param sdkconfig_name: Filter tests whose apps are built with this sdkconfig
            name
        """
        self.cli_target = cli_target
        self.sdkconfig_name = sdkconfig_name

        settings = CiSettings()
        if settings.is_in_ci:
            self.apps = settings.get_built_apps_list()
        else:
            self.apps = None

        self.cases: t.List[PytestCase] = []

    @staticmethod
    def get_case_by_item(item: pytest.Item) -> t.Optional[PytestCase]:
        """Get the test case associated with a pytest item.

        :param item: The pytest test item

        :returns: PytestCase object or None if not found
        """
        return item.stash.get(IDF_CI_PYTEST_CASE_KEY, None)

    @pytest.fixture
    @multi_dut_argument
    def target(
        self,
        request: FixtureRequest,
    ) -> str:
        """Fixture that provides the target for tests.

        :param request: Pytest fixture request

        :returns: Target string

        :raises ValueError: If target parameter is not defined
        """
        target_value = getattr(request, 'param', None)
        if not target_value:
            raise ValueError('"target" must be defined in pytest.mark.parametrize')
        return target_value

    @pytest.fixture
    @multi_dut_argument
    def config(self, request: FixtureRequest) -> str:
        """Fixture that provides the configuration for tests.

        :param request: Pytest fixture request

        :returns: Configuration string, defaults to 'default' if not specified
        """
        return getattr(request, 'param', None) or 'default'

    @pytest.fixture
    @multi_dut_fixture
    def build_dir(
        self,
        request: FixtureRequest,
        app_path: str,
        target: t.Optional[str],
        config: t.Optional[str],
    ) -> str:
        """Find a valid build directory based on priority rules.

        Checks local build directories in the following order:

        1. build_<target>_<config>
        2. build_<target>
        3. build_<config>
        4. build

        :param request: Pytest fixture request
        :param app_path: Path to the application
        :param target: Target being used
        :param config: Configuration being used

        :returns: Valid build directory name

        :raises ValueError: If no valid build directory is found
        """
        check_dirs = []
        build_dir_arg = request.config.getoption('build_dir')

        if build_dir_arg:
            check_dirs.append(build_dir_arg)
        if target is not None and config is not None:
            check_dirs.append(f'build_{target}_{config}')
        if target is not None:
            check_dirs.append(f'build_{target}')
        if config is not None:
            check_dirs.append(f'build_{config}')
        check_dirs.append('build')

        for check_dir in check_dirs:
            binary_path = os.path.join(app_path, check_dir)
            if os.path.isdir(binary_path):
                logger.info(f'Found valid binary path: {binary_path}')
                return check_dir

            logger.warning('Checking binary path: %s... missing... trying another location', binary_path)

        raise ValueError(
            'No valid build directory found. '
            f'Please build the binary via "idf.py -B {check_dirs[0]} build" and run pytest again'
        )

    @pytest.hookimpl(tryfirst=True)
    def pytest_pycollect_makemodule(
        self,
        module_path: Path,
    ):
        """Handle module collection for pytest, mocking any missing modules.

        This hook runs before module collection to prevent errors from missing
        dependencies by automatically mocking them.

        :param module_path: Path to the module being collected
        """
        while True:
            try:
                spec = importlib.util.spec_from_file_location('', module_path)
                if spec:
                    module = importlib.util.module_from_spec(spec)
                    if spec.loader and module:
                        spec.loader.exec_module(module)
                break
            except ModuleNotFoundError as e:
                match = _MODULE_NOT_FOUND_REGEX.search(e.msg)
                if match:
                    pkg = match.group(1)
                    logger.warning('Mocking missing package during collection: %s', pkg)
                    sys.modules[pkg] = MagicMock()
                    continue

    @pytest.hookimpl(wrapper=True)
    def pytest_collection_modifyitems(self, config: Config, items: t.List[Function]):
        """Filter test cases based on target, sdkconfig, and available apps.

        :param config: Pytest configuration
        :param items: Collected test items
        """
        # Create PytestCase objects for all items
        for item in items:
            item.stash[IDF_CI_PYTEST_CASE_KEY] = PytestCase.from_item(item)
            item.stash[IDF_CI_PYTEST_DEBUG_INFO_KEY] = dict()

        # Add markers to items
        for item in items:
            case = self.get_case_by_item(item)
            if case is None:
                continue

            # Add 'host_test' marker to host test cases
            if 'qemu' in case.all_markers or 'linux' in case.targets:
                item.add_marker(pytest.mark.host_test)

        yield

        deselected_items: t.List[Function] = []

        # Filter by nightly_run marker
        if os.getenv('INCLUDE_NIGHTLY_RUN') == '1':
            # Include both nightly_run and non-nightly_run cases
            pass
        else:  # Determine if we should select nightly_run tests or non-nightly_run tests
            nightly_or_not = os.getenv('NIGHTLY_RUN') == '1'
            filtered_items = []
            for item in items:
                case = self.get_case_by_item(item)
                if case is None:
                    continue

                if ('nightly_run' in case.all_markers) == nightly_or_not:
                    filtered_items.append(item)
                else:
                    deselected_items.append(item)
            items[:] = filtered_items

        # Filter by target
        if self.cli_target != 'all':
            filtered_items = []
            for item in items:
                case = self.get_case_by_item(item)
                if case is None:
                    continue

                if case.target_selector != self.cli_target:
                    item.add_marker(pytest.mark.skip(reason=f'Target mismatch: {self.cli_target}'))
                    deselected_items.append(item)
                else:
                    filtered_items.append(item)
            items[:] = filtered_items

        # Filter by sdkconfig_name
        if self.sdkconfig_name:
            filtered_items = []
            for item in items:
                case = self.get_case_by_item(item)
                if case is None:
                    continue

                if self.sdkconfig_name not in set(app.config for app in case.apps):
                    item.stash[IDF_CI_PYTEST_DEBUG_INFO_KEY]['skip_reason'] = (
                        f'sdkconfig name mismatch. '
                        f'app sdkconfigs: {case.configs}, but CLI specified: {self.sdkconfig_name}'
                    )
                    deselected_items.append(item)
                else:
                    filtered_items.append(item)
            items[:] = filtered_items

        # Filter by app list
        if self.apps is not None:
            app_dirs = [os.path.abspath(app.build_path) for app in self.apps]
            filtered_items = []
            for item in items:
                case = self.get_case_by_item(item)
                if case is None:
                    continue

                skip_reason = case.get_skip_reason_if_not_built(app_dirs)
                if skip_reason:
                    item.stash[IDF_CI_PYTEST_DEBUG_INFO_KEY]['skip_reason'] = skip_reason
                    deselected_items.append(item)
                else:
                    filtered_items.append(item)
            items[:] = filtered_items

        # Report deselected items
        config.hook.pytest_deselected(items=deselected_items)

    def pytest_report_collectionfinish(self, items: t.List[Function]) -> None:
        for item in items:
            case = self.get_case_by_item(item)
            if case is None:
                continue

            self.cases.append(case)


##################
# Hook Functions #
##################
def pytest_addoption(parser: pytest.Parser):
    """Add custom command line options for IDF pytest plugin.

    :param parser: Pytest command line parser
    """
    # CLI values
    idf_ci_group = parser.getgroup('idf_ci')
    idf_ci_group.addoption(
        '--sdkconfig',
        help='Run only tests whose apps are built with this sdkconfig name',
    )

    # INI values
    parser.addini(
        'env_markers',
        help='Markers that indicate the running environment of the test case. '
        'Each line is a `<marker_name>: <marker_description>` pair',
        type='linelist',
    )


def pytest_configure(config: Config):
    """Configure the pytest environment for IDF tests.

    :param config: Pytest configuration object
    """
    setup_logging(config.getoption('log_cli_level'))

    cli_target = config.getoption('target') or 'all'
    sdkconfig_name = config.getoption('sdkconfig')

    env_markers: t.Set[str] = set()
    for line in config.getini('env_markers'):
        name, _ = line.split(':', maxsplit=1)
        config.addinivalue_line('markers', line)
        env_markers.add(name)

    PytestCase.KNOWN_ENV_MARKERS = env_markers

    plugin = IdfPytestPlugin(cli_target=cli_target, sdkconfig_name=sdkconfig_name)
    config.stash[IDF_CI_PLUGIN_KEY] = plugin
    config.pluginmanager.register(plugin)

    # Add markers definitions
    config.addinivalue_line('markers', 'host_test: this test case runs on host machines')
    config.addinivalue_line('markers', 'nightly_run: this test case is a nightly run')


def pytest_unconfigure(config: Config):
    """Clean up the IDF pytest plugin when pytest is shutting down.

    :param config: Pytest configuration object
    """
    plugin = config.stash.get(IDF_CI_PLUGIN_KEY, None)
    if plugin:
        del config.stash[IDF_CI_PLUGIN_KEY]
        config.pluginmanager.unregister(plugin)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import typing as t

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict  # noqa

PathLike = t.Union[str, os.PathLike]


class UndefinedType:
    def __repr__(self) -> str:
        return '__UNDEF__'


UNDEF = UndefinedType()
T = t.TypeVar('T')
UndefinedOr = t.Union[UndefinedType, T]


def is_undefined(value: t.Any) -> bool:
    """Check if a value is undefined.

    :param value: The value to check

    :returns: True if the value equals the UNDEF constant, False otherwise
    """
    return isinstance(value, UndefinedType) or value == '__UNDEF__'  # click would convert UNDEF to '__UNDEF__'


def is_defined_and_satisfies(value: t.Any, _callable: t.Callable[[t.Any], bool] = bool) -> bool:
    """Check if a value is defined and the callable returns True for the value.

    :param value: The value to check
    :param _callable: The callable to check the value with. Default is ``bool()``

    :returns: True if the value is defined and the callable returns True for the value,
        False otherwise
    """
    return not is_undefined(value) and _callable(value)



# This file contains snippets from various files in the repository. Attached the LICENSE file accordingly.
import functools
import os
import re

# The `glob.translate` function is introduced in python 3.13.
# Copying the implementation from the Python 3.13 source code.


# This `translate` function is taken from
# https://github.com/python/cpython/blob/2c8f329dc634290fb88636f85c05e473bc0104d5/Lib/glob.py#L267
# following the license terms of the Python Software Foundation License Version 2.
def translate(pat, *, recursive=False, include_hidden=False, seps=None):
    """Translate a pathname with shell wildcards to a regular expression.

    If `recursive` is true, the pattern segment '**' will match any number of path
    segments.

    If `include_hidden` is true, wildcards can match path segments beginning with a dot
    ('.').

    If a sequence of separator characters is given to `seps`, they will be used to split
    the pattern into segments and match path separators. If not given, os.path.sep and
    os.path.altsep (where available) are used.
    """
    if not seps:
        if os.path.altsep:
            seps = (os.path.sep, os.path.altsep)
        else:
            seps = os.path.sep
    escaped_seps = ''.join(map(re.escape, seps))
    any_sep = f'[{escaped_seps}]' if len(seps) > 1 else escaped_seps
    not_sep = f'[^{escaped_seps}]'
    if include_hidden:
        one_last_segment = f'{not_sep}+'
        one_segment = f'{one_last_segment}{any_sep}'
        any_segments = f'(?:.+{any_sep})?'
        any_last_segments = '.*'
    else:
        one_last_segment = f'[^{escaped_seps}.]{not_sep}*'
        one_segment = f'{one_last_segment}{any_sep}'
        any_segments = f'(?:{one_segment})*'
        any_last_segments = f'{any_segments}(?:{one_last_segment})?'

    results = []
    parts = re.split(any_sep, pat)
    last_part_idx = len(parts) - 1
    for idx, part in enumerate(parts):
        if part == '*':
            results.append(one_segment if idx < last_part_idx else one_last_segment)
        elif recursive and part == '**':
            if idx < last_part_idx:
                if parts[idx + 1] != '**':
                    results.append(any_segments)
            else:
                results.append(any_last_segments)
        else:
            if part:
                if not include_hidden and part[0] in '*?':
                    results.append(r'(?!\.)')
                results.extend(_fnmatch_translate(part, f'{not_sep}*', not_sep)[0])
            if idx < last_part_idx:
                results.append(any_sep)
    res = ''.join(results)
    return rf'(?s:{res})\Z'


# This `_re_setops_sub` function is taken from
# https://github.com/python/cpython/blob/2c8f329dc634290fb88636f85c05e473bc0104d5/Lib/fnmatch.py#L83
# following the license terms of the Python Software Foundation License Version 2.
_re_setops_sub = re.compile(r'([&~|])').sub

# This `_re_escape` function is taken from
# https://github.com/python/cpython/blob/2c8f329dc634290fb88636f85c05e473bc0104d5/Lib/fnmatch.py#L84
# following the license terms of the Python Software Foundation License Version 2.
_re_escape = functools.lru_cache(maxsize=512)(re.escape)


# This `_fnmatch_translate` function is taken from
# https://github.com/python/cpython/blob/2c8f329dc634290fb88636f85c05e473bc0104d5/Lib/fnmatch.py#L86
# following the license terms of the Python Software Foundation License Version 2.
def _fnmatch_translate(pat, star, question_mark):
    res = []
    add = res.append
    star_indices = []

    i, n = 0, len(pat)
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '*':
            # store the position of the wildcard
            star_indices.append(len(res))
            add(star)
            # compress consecutive `*` into one
            while i < n and pat[i] == '*':
                i += 1
        elif c == '?':
            add(question_mark)
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                add('\\[')
            else:
                stuff = pat[i:j]
                if '-' not in stuff:
                    stuff = stuff.replace('\\', r'\\')
                else:
                    chunks = []
                    k = i + 2 if pat[i] == '!' else i + 1
                    while True:
                        k = pat.find('-', k, j)
                        if k < 0:
                            break
                        chunks.append(pat[i:k])
                        i = k + 1
                        k = k + 3
                    chunk = pat[i:j]
                    if chunk:
                        chunks.append(chunk)
                    else:
                        chunks[-1] += '-'
                    # Remove empty ranges -- invalid in RE.
                    for k in range(len(chunks) - 1, 0, -1):
                        if chunks[k - 1][-1] > chunks[k][0]:
                            chunks[k - 1] = chunks[k - 1][:-1] + chunks[k][1:]
                            del chunks[k]
                    # Escape backslashes and hyphens for set difference (--).
                    # Hyphens that create ranges shouldn't be escaped.
                    stuff = '-'.join(s.replace('\\', r'\\').replace('-', r'\-') for s in chunks)
                i = j + 1
                if not stuff:
                    # Empty range: never match.
                    add('(?!)')
                elif stuff == '!':
                    # Negated empty range: match any character.
                    add('.')
                else:
                    # Escape set operations (&&, ~~ and ||).
                    stuff = _re_setops_sub(r'\\\1', stuff)
                    if stuff[0] == '!':
                        stuff = '^' + stuff[1:]
                    elif stuff[0] in ('^', '['):
                        stuff = '\\' + stuff
                    add(f'[{stuff}]')
        else:
            add(_re_escape(c))
    assert i == n
    return res, star_indices



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import textwrap
from pathlib import Path

import pytest
from conftest import create_project

from idf_ci.scripts import build


@pytest.mark.skipif(not os.getenv('IDF_PATH'), reason='IDF_PATH is not set')
class TestBuild:
    @pytest.fixture(autouse=True)
    def setup_test_projects(self, tmp_path: Path):
        """tmp_path/foo
            - test_foo.py
                  - test_foo_slow (-m slow)

        tmp_path/bar
            - test_bar.py
                  - test_bar_slow (-m slower)

        - filter_expr='slow' should match both test_foo_slow and test_bar_slow
        - marker_expr='slow' should match only test_foo_slow
        - marker_expr='slower' should match only test_bar_slow
        """

        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        with open(tmp_path / 'foo' / 'test_foo.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32',
            ], indirect=True)
            @pytest.mark.slow
            def test_foo_slow(dut):
                pass
            """)
            )

        with open(tmp_path / 'bar' / 'test_bar.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32',
            ], indirect=True)
            @pytest.mark.slower
            def test_bar_slow(dut):
                pass
            """)
            )

        self.paths = [str(tmp_path / 'foo'), str(tmp_path / 'bar')]
        self.tmp_path = tmp_path

    def test_build_with_filter_or_marker_sets_only_test_related(self) -> None:
        # Test: Build without filter_expr and without only_test_related
        built_apps, _ = build(
            paths=self.paths,
            dry_run=True,
        )
        assert len(built_apps) > 2

        # Test: Build with filter_expr
        built_apps, _ = build(
            paths=self.paths,
            filter_expr='slow',
            dry_run=True,
        )
        assert len(built_apps) == 2

        # Test: Build with marker_expr
        built_apps, _ = build(
            paths=self.paths,
            marker_expr='slow',
            dry_run=True,
        )
        assert len(built_apps) == 1

        built_apps, _ = build(
            paths=self.paths,
            marker_expr='slower',
            dry_run=True,
        )
        assert len(built_apps) == 1

        # Test: Build with filter_expr and explicitly set only_test_related=False
        built_apps, _ = build(
            paths=self.paths,
            filter_expr='slow',
            only_test_related=False,
            dry_run=True,
        )
        assert len(built_apps) == 2

        # Test: Build with marker_expr and explicitly set only_test_related=False
        built_apps, _ = build(
            paths=self.paths,
            marker_expr='slow',
            only_test_related=False,
            dry_run=True,
        )
        assert len(built_apps) == 1

        # Test: Build with only_test_related=True but no filter_expr
        built_apps, _ = build(
            paths=self.paths,
            only_test_related=True,
            dry_run=True,
        )
        assert len(built_apps) == 2

        # Test: Build with only_non_test_related=True
        # Should build only non-test-related apps
        built_apps, _ = build(
            paths=self.paths,
            target='esp32',
            only_non_test_related=True,
            dry_run=True,
        )

        # Should build only non-test-related apps (none for esp32 since foo and bar are test-related)
        assert len(built_apps) == 0



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import textwrap
from pathlib import Path

import pytest

from idf_ci import get_pytest_cases
from idf_ci.cli import click_cli


class TestGetPytestCases:
    @pytest.fixture(autouse=True)
    def _setup(self, runner):
        assert runner.invoke(click_cli, ['test', 'init']).exit_code == 0

        yield

    TEMPLATE_SCRIPT = textwrap.dedent("""
        import pytest

        @pytest.mark.parametrize('target', [
            'esp32',
            'esp32c3',
        ], indirect=True)
        def test_foo_single(dut):
            pass

        @pytest.mark.parametrize('count,target', [
            (2, 'esp32|esp32s2'),
            (3, 'esp32s2|esp32s2|esp32s3'),
        ], indirect=True)
        def test_foo_multi(dut):
            pass

        @pytest.mark.parametrize('target', [
            'linux',
        ], indirect=True)
        def test_foo_host(dut):
            pass

        @pytest.mark.parametrize('target', [
            'esp32',
        ], indirect=True)
        @pytest.mark.qemu
        def test_foo_qemu(dut):
            pass
        """)

    def test_get_single_specific(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_get_single_specific.py'
        script.write_text(self.TEMPLATE_SCRIPT)
        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32')

        assert len(cases) == 1
        assert cases[0].targets == ['esp32']
        assert cases[0].name == 'test_foo_single'

    def test_get_multi_specific(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_get_multi_specific.py'
        script.write_text(self.TEMPLATE_SCRIPT)
        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32s2,esp32s2,esp32s3')

        assert len(cases) == 1
        assert cases[0].targets == ['esp32s2', 'esp32s2', 'esp32s3']
        assert cases[0].name == 'test_foo_multi'

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32s3,esp32s2,esp32s2')  # order matters
        assert len(cases) == 0

    def test_get_by_filter(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_get_by_filter.py'
        script.write_text(self.TEMPLATE_SCRIPT)
        cases = get_pytest_cases(paths=[str(tmp_path)], target='linux', filter_expr='foo')

        assert len(cases) == 1
        assert cases[0].name == 'test_foo_host'

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32', filter_expr='foo')
        assert len(cases) == 1
        assert cases[0].name == 'test_foo_single'

    def test_get_all(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_get_multi_all.py'
        script.write_text(self.TEMPLATE_SCRIPT)
        cases = get_pytest_cases(paths=[str(tmp_path)])

        assert len(cases) == 4
        assert cases[0].name == 'test_foo_single'
        assert cases[0].targets == ['esp32']

        assert cases[1].name == 'test_foo_single'
        assert cases[1].targets == ['esp32c3']

        assert cases[2].name == 'test_foo_multi'
        assert cases[2].targets == ['esp32', 'esp32s2']

        assert cases[3].name == 'test_foo_multi'
        assert cases[3].targets == ['esp32s2', 'esp32s2', 'esp32s3']

    def test_filter_with_sdkconfig_name(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_filter_with_sdkconfig_name.py'
        script.write_text(
            textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('config,target', [
                ('foo', 'esp32'),
                ('bar', 'esp32'),
            ], indirect=True)
            def test_filter_with_sdkconfig_name_single_dut(dut, config):
                pass

            @pytest.mark.parametrize('count', [2], indirect=True)
            @pytest.mark.parametrize('config,target', [
                ('foo|bar', 'esp32'),
                ('bar|baz', 'esp32'),
            ], indirect=True)
            def test_filter_with_sdkconfig_name_multi_dut(dut, config):
                pass
            """)
        )

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32', sdkconfig_name='foo')
        assert len(cases) == 1
        assert cases[0].caseid == 'esp32.foo.test_filter_with_sdkconfig_name_single_dut'
        assert cases[0].apps[0].build_dir == str(tmp_path / 'build_esp32_foo')

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32,esp32', sdkconfig_name='foo')
        assert len(cases) == 1

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32,esp32', sdkconfig_name='bar')
        assert len(cases) == 2

    def test_host_test(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_host_test.py'
        script.write_text(self.TEMPLATE_SCRIPT)

        cases = get_pytest_cases(paths=[str(tmp_path)], target='linux')
        assert len(cases) == 1
        assert cases[0].name == 'test_foo_host'

        cases = get_pytest_cases(paths=[str(tmp_path)], target='all', marker_expr='host_test')
        assert len(cases) == 2
        assert cases[0].name == 'test_foo_host'
        assert cases[1].name == 'test_foo_qemu'

    def test_custom_app_path(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_custom_app_path.py'
        script.write_text(
            textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('count, app_path, target, config', [
                (3, None, 'esp32s2', None),
                (2, 'subdir', 'esp32s3', 'foo'),
            ], indirect=True)
            def test_multi_dut_with_custom_app_path(dut, config):
                pass
            """)
        )

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32s3,esp32s3')
        assert len(cases) == 1
        assert cases[0].caseid == "('esp32s3', 'esp32s3').('foo', 'foo').test_multi_dut_with_custom_app_path"
        assert cases[0].apps[0].build_dir == str(tmp_path / 'subdir' / 'build_esp32s3_foo')
        assert cases[0].apps[1].build_dir == str(tmp_path / 'subdir' / 'build_esp32s3_foo')

    def test_no_params(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_no_params.py'
        script.write_text(
            textwrap.dedent("""
            def test_no_param():
                pass
            """)
        )

        cases = get_pytest_cases(paths=[str(tmp_path)])
        assert len(cases) == 0  # since target is required

    def test_qemu_caseid(self, tmp_path: Path) -> None:
        script = tmp_path / 'test_qemu_caseid.py'
        script.write_text(self.TEMPLATE_SCRIPT)

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32', marker_expr='qemu')
        assert len(cases) == 1
        assert cases[0].name == 'test_foo_qemu'
        assert cases[0].caseid == 'esp32_qemu.default.test_foo_qemu'

        cases = get_pytest_cases(paths=[str(tmp_path)], target='esp32')
        assert len(cases) == 1
        assert cases[0].name == 'test_foo_single'
        assert cases[0].caseid == 'esp32.default.test_foo_single'



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

__all__ = [
    'CiSettings',
    'GitlabEnvVars',
    'IdfPytestPlugin',
    'PytestApp',
    'PytestCase',
    'build',
    'get_all_apps',
    'get_pytest_cases',
]


from idf_ci.envs import GitlabEnvVars
from idf_ci.idf_pytest import IdfPytestPlugin, PytestApp, PytestCase, get_pytest_cases
from idf_ci.scripts import build, get_all_apps
from idf_ci.settings import CiSettings
